#!/usr/bin/env python

##############################################################################
##
## Copyright (C) 2021 Narda Safety Test Solutions GmbH.
## Contact: www.narda-sts.com
##
## Redistribution and use in source and binary forms,
## with or without modification, are permitted provided that the following
## conditions are met:
##
## 1.) Redistributions of source code must retain the above copyright notice,
## this list of conditions and the following disclaimer.
##
## 2.) Redistributions in binary form must reproduce the above copyright
## notice, this list of conditions and the following disclaimer in the
## documentation and/or other materials provided with the distribution.
##
## 3.) Neither the name of the copyright holder nor the names of its
## contributors may be used to endorse or promote products derived from this
## software without specific prior written permission.
##
## 4.) EXPORT CONTROL
## The Software, including technical data /cryptographic software,
## may be subject to, German, European Union and U.S. export controls and
## may be subject to import or export controls in other countries.
## The Licensee agrees to strictly comply with all applicable import and
## export regulations. He specifically agrees, that he must not disclose or
## otherwise export or re-export the Licensed Software or any part thereof
## delivered under this end user license agreement (EULA) to any country
## (including a national or resident of such country) without a valid export
## or import license. Please be aware that the Software may contain
## US-Content, therefor the Licensee represent and warrant that he is not a
## citizen of, or otherwise located within, an embargoed nation (including
## without limitation Iran, Syria, Sudan, Cuba, North Korea) and that he is
## not otherwise prohibited under the Export Laws from receiving the Software.
## All rights to Use the Software are granted on condition that such rights
## are forfeited if the Licensee fails to comply with
## the terms of this Agreement.
##
## 5.) SEVERABILITY
## Should any provision of this Agreement be or become invalid, ineffective
## or unenforceable, the remaining provisions of this Agreement shall be
## valid. The parties agree to replace the invalid, ineffective or
## unenforceable provision by a valid, effective and enforceable provision
## which best meets the commercial intention of the parties.
## The same shall apply in case of omissions.
##
## 6.) APPLICABLE LAW AND PLACE OF JURISTICATION
## This Agreement shall be constituted under the law of Germany.
## The United Nations Convention on the International Sale of Goods
## shall not apply to this Agreement. The Place of Jurisdiction for any
## dispute between the Parties shall be Tübingen (Germany).
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
## “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
## TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Narda Safety Test Solutions GmbH
## BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
## CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
## SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
## INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
## CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
## ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
## POSSIBILITY OF SUCH DAMAGE.
##
##############################################################################


from pathlib import Path
import time
import os
from math import fabs
import wx
from nardascripting.base.usrscriptbase import *
from nardascripting.base.signalsharkdev import *
from nardascripting.base.measthread import *
from . import geocalc
from .cmapsettings import CMapSettings
from .gui.cmapdlgs import LevelData, MeasParameters, MapDlg, CMapSettingsDlg, check_map_tiles
import csv
import datetime as dt


APP_TITLE ='Coverage Map 1.0.0'


class CMap(UsrScriptBase):
    """User script class"""

    def __init__(self, main_gui, dev=SignalSharkDev()):
        """Initialization. Please leave this code unchanged"""
        super().__init__(main_gui, dev, __file__)

        # Base settings
        self._tab_name = 'nslcoveragemap'
        self._scr_title = APP_TITLE
        self._scr_description = 'Measures the level by location and displays the values on a map.'
        # self._icon_path = self.script_path.joinpath('NardaScriptLauncher_icon_close.png')
        self._list_prio = 0
        self._nsl_executed_behavior = NSL_Executed_Behaviors.MINIMIZE_NSL

        # Add class variables if needed
        # -------------------------------------------------------------------------------
        self.settings = {}

    def _run_script(self, args):
        """
        Script main function
        This method is called when a user clicks on the corresponding script button.
        """
        # GNSS settings hint
        # -------------------------------------------------------------------------------
        if self.MessageBoxModal('If you are using the internal GNSS receiver of SignalShark, please check the ' +
                                '"Dyn. Platform Model for internal GNSS" parameter in the SignalShark settings menu. ' +
                                'This should be set to "Automotive". \r\n\r\nIs the setting correct?',
                                'GNSS Settings Check', wx.YES_NO | wx.ICON_QUESTION) == wx.NO:
            return

        # Load coverage map settings
        # -------------------------------------------------------------------------------
        self.settings = CMapSettings.load()
        if not self.settings:
            self.MessageBoxModal('Please run App Coverage Map Settings first!',
                                 'Coverage Map', wx.OK | wx.ICON_INFORMATION)
            return

        # -------------------------------------------------------------------------------
        if not check_map_tiles(self.settings.get('map_path')):
            self.MessageBoxModal('Maps not found. Please re-run the settings to select the '
                                 '\'Mapnik\' folder in the Maps sub-directory.',
                                 'Map Files Error', wx.OK | wx.ICON_ERROR)
            return

        # Connect to the device
        # -------------------------------------------------------------------------------
        # Connect to the device
        if not self.signalshark.connect():
            self.MessageBoxModal('Cannot connect to the device!', 'Connection error',
                                 wx.OK | wx.ICON_ERROR)
            return
        self.signalshark.scpi.check_error()

        # Check GNSS:
        gnss_data = self.signalshark.scpi.sense.get_gnss_data()
        if gnss_data.quality == GnssQuality.NOFIX or gnss_data.quality == GnssQuality.FROZEN:
            self.MessageBoxModal('No GNSS reception! Please check the GNSS settings and the GNSS antenna if necessary.',
                                 'GNSS',
                                 wx.OK | wx.ICON_ERROR)
            self.signalshark.disconnect()
            return

        # Setup Level Meter:
        # -------------------------------------------------------------------------------
        # Add a RT Spectrum Task if it does not already exist.
        if not self.signalshark.scpi.check_add_task(TaskTypes.RT_SPECTRUM):
            self.MessageBoxModal('Please configure the newly added task (e.g. frequency range)', 'New Task',
                                 wx.OK | wx.ICON_INFORMATION)
            self.signalshark.disconnect()
            return

        # Add a Level Meter View if it does not already exist and configure it
        exists, view_index = self.signalshark.scpi.check_add_view(ViewTypes.LEVEL)
        if view_index < 1:
            self.MessageBoxModal('Adding Level Meter View not possible.',
                                 'Level Meter View', wx.OK | wx.ICON_ERROR)
            self.signalshark.disconnect()
            return

        if not exists:
            self.MessageBoxModal('Please configure the newly added Level Meter View (e.g. Ftune, CBW)',
                                 'Level Meter View', wx.OK | wx.ICON_INFORMATION)
            self.signalshark.disconnect()
            return

        self.meas_param = MeasParameters()

        # Set Unit
        self.signalshark.scpi.check_error()
        self.meas_param.unit = LevelUnits.from_string(self.settings.get('unit', 'dBm'))
        self.signalshark.scpi.display.set_unit(self.meas_param.unit)
        test_unit = self.signalshark.scpi.display.get_unit()
        if test_unit != self.meas_param.unit:
            self.MessageBoxModal('This unit can only be set, '
                                 'if a Narda antenna is connected or a user antenna is selected.',
                                 'Unit error', wx.OK | wx.ICON_INFORMATION)
            self.signalshark.disconnect()
            return

        # Set Level Meter detector
        self.meas_param.det = LevDetectorModes.from_string(self.settings.get('cbx_detector', 'RMS'))
        det_num = 0
        if self.meas_param.det == LevDetectorModes.PPk or self.meas_param.det == LevDetectorModes.CPk:
            det_num = 1
        elif self.meas_param.det == LevDetectorModes.RMS or self.meas_param.det == LevDetectorModes.CRMS:
            det_num = 2
        else:
            det_num = 3
        self.signalshark.scpi.levelmeter.set_detector(det_num, self.meas_param.det)

        # Ask for comment:
        # -------------------------------------------------------------------------------
        self.meas_param.comment = ''
        dlg = wx.TextEntryDialog(self.frm_dlghelper, message='Optional text comment:', caption='Comment')
        if self.ShowDlgModalTop(dlg) == wx.ID_OK:
            self.meas_param.comment = dlg.GetValue()
        dlg.Destroy()  # Please do not forget to destroy the created dialog if you do not need it anymore

        # Setup measurement storage file:
        # -------------------------------------------------------------------------------
        time_stamp = dt.datetime.now()
        base_path = self.settings.get('dirpck_file_path', r'c:\\')
        filename_prefix = self.settings.get('tctr_file_prefix', 'Measurement')
        file_name = time_stamp.strftime(filename_prefix + '_%Y-%m-%d_%H-%M-%S.csv')
        file_path = Path(base_path).joinpath(file_name)

        # Get measurement parameters:
        # -------------------------------------------------------------------------------
        self.meas_param.input = self.signalshark.scpi.sense.get_input()
        self.meas_param.atten = self.signalshark.scpi.sense.get_attenuator()
        self.meas_param.ftune = self.signalshark.scpi.levelmeter.get_frequency_tune()
        self.meas_param.cbw = self.signalshark.scpi.levelmeter.get_cbw()
        self.meas_param.meas_time = self.signalshark.scpi.levelmeter.get_measurement_time()
        self.meas_param.post_avg = self.signalshark.scpi.levelmeter.get_post_avg()
        self.meas_param.pre_avg = self.signalshark.scpi.levelmeter.get_pre_avg()

        # Disconnect from the device
        self.signalshark.disconnect()


        # Add map display dialog.
        # -------------------------------------------------------------------------------
        map_dlg = MapDlg(self.main_gui, APP_TITLE, self.settings, str(file_path),
                         self.meas_param)
        point_dist = self.settings.get('point_dist', 500)

        #  Run measurement in additional task to keep GUI alive.
        # -------------------------------------------------------------------------------
        mthread = MeasDlgThread(main_gui=self.main_gui, dlg=map_dlg,
                                callback=CMap._run_measurement,
                                args=[self.signalshark.addr, str(file_path),
                                      point_dist, det_num, self.meas_param.clone()])

        # Start the measurement thread and show a visualization (progress) dialog.
        mthread.start_measurement()

    @staticmethod
    def _run_measurement(stopevent, update_progress, wait_for_stopevent,
                         signalshark_addr: str, file_name: str,
                         point_dist: float, det_num: int, meas_param: MeasParameters):
        """Main method of the measurement thread that handles a time consuming measurement.

        :param stopevent: Thread flag that indicates whether the procedure should be finished/cancelled
        :param update_progress: Delegate method to update a progress bar, message string and icon
        :param wait_for_stopevent: Delegate method to wait until stopevent is raised with possibility to show a message.
        :param signalshark_addr: Custom parameter to handover the ip address of the SignalShark
        :param file_name: Custom parameter to handover the file name
        :param point_dist: Custom parameter to handover the distance between two measurement points
        :param det_num: Custom parameter to handover the detector number
        :param meas_param: Custom parameter to handover the measurement parameters
        :return:
        """

        # Setup connection to the SignalShark and display error message if connection fails.
        # -------------------------------------------------------------------------------
        signalshark = SignalSharkDev(signalshark_addr)
        if not signalshark.connect():
            return wait_for_stopevent('Cannot connect to the device!', icon_style=wx.ICON_ERROR)

        data_file = None
        try:
            # Update the dialog message text and show a progress bar
            # -------------------------------------------------------------------------------
            update_progress(msg='Press Quit to stop measurement.',
                            show_progress=True, btn_style= wx.OK)

            # Create file and file header:
            data_file = open(file_name, 'a', newline='')
            data_writer = csv.writer(data_file, delimiter=';', quotechar='\'', quoting=csv.QUOTE_MINIMAL)
            data_writer.writerow(['input', meas_param.input])
            data_writer.writerow(['atten', meas_param.atten])
            data_writer.writerow(['unit', meas_param.unit])
            data_writer.writerow(['ftune', meas_param.ftune])
            data_writer.writerow(['cbw', meas_param.cbw])
            data_writer.writerow(['det', meas_param.det])
            data_writer.writerow(['meas_time', meas_param.meas_time])
            data_writer.writerow(['post_avg', meas_param.post_avg])
            data_writer.writerow(['pre_avg', meas_param.pre_avg])
            data_writer.writerow(['comment', meas_param.comment])

            # Reserved lines
            for i in range(7):
                data_writer.writerow([])

            data_writer.writerow(['tstamp', 'lat', 'lon', 'gnss_fix', meas_param.det])

            # Do level measurements
            # -------------------------------------------------------------------------------
            check_cnt = 0
            last_pos = GnssCoordinate()
            previous_scan_number = 0
            data_lst = []
            while True:
                # Periodically check, whether the user has clicked the cancel button,
                # and exit thread if so.
                if check_cnt >= 10:
                    if stopevent.is_set():
                        return
                    check_cnt = 0
                    data_file.flush()
                    if len(data_lst) > 0:
                        if update_progress(data=data_lst):
                            data_lst = []

                # Wait until Scan Number has increased.
                scan_number = signalshark.scpi.levelmeter.get_data_update(previous_scan_number)
                if scan_number < 0:
                    return wait_for_stopevent('Cannot acquire measurement data!', icon_style=wx.ICON_ERROR)
                if previous_scan_number >= scan_number:
                    continue
                previous_scan_number = scan_number

                # Get measurement:
                gnss_data = signalshark.scpi.sense.get_gnss_data()
                if geocalc.check_gnss(gnss_data):
                    meas_pos = GnssCoordinate(gnss_data.latitude, gnss_data.longitude)
                    distance = fabs(geocalc.get_distance_haversine(last_pos, meas_pos))
                    if distance >= point_dist:
                        timestamp = signalshark.scpi.levelmeter.get_tstamp()
                        lev, maxlev = signalshark.scpi.levelmeter.get_data_detector(det_num)
                        if timestamp and lev:
                            last_pos = GnssCoordinate(meas_pos.latitude, meas_pos.longitude)
                            data = LevelData()
                            data.timestamp = timestamp
                            data.gnss = meas_pos
                            data.det_val = lev
                            data.unit = meas_param.unit
                            data_writer.writerow([timestamp.datetime, meas_pos.latitude, meas_pos.longitude,
                                                  gnss_data.fix, lev])
                            data_lst.append(data)
                check_cnt += 1
                # Allow the thread to synchronize the stop event.
                stopevent.wait(0.05)
        except Exception as e:
            # Do some error handling
            return wait_for_stopevent(str(e), icon_style=wx.ICON_ERROR)
        finally:
            # Close file and connection.
            if data_file:
                data_file.close()
            signalshark.disconnect()
