#!/usr/bin/env python

##############################################################################
##
## Copyright (C) 2021 Narda Safety Test Solutions GmbH.
## Contact: www.narda-sts.com
##
## Redistribution and use in source and binary forms,
## with or without modification, are permitted provided that the following
## conditions are met:
##
## 1.) Redistributions of source code must retain the above copyright notice,
## this list of conditions and the following disclaimer.
##
## 2.) Redistributions in binary form must reproduce the above copyright
## notice, this list of conditions and the following disclaimer in the
## documentation and/or other materials provided with the distribution.
##
## 3.) Neither the name of the copyright holder nor the names of its
## contributors may be used to endorse or promote products derived from this
## software without specific prior written permission.
##
## 4.) EXPORT CONTROL
## The Software, including technical data /cryptographic software,
## may be subject to, German, European Union and U.S. export controls and
## may be subject to import or export controls in other countries.
## The Licensee agrees to strictly comply with all applicable import and
## export regulations. He specifically agrees, that he must not disclose or
## otherwise export or re-export the Licensed Software or any part thereof
## delivered under this end user license agreement (EULA) to any country
## (including a national or resident of such country) without a valid export
## or import license. Please be aware that the Software may contain
## US-Content, therefor the Licensee represent and warrant that he is not a
## citizen of, or otherwise located within, an embargoed nation (including
## without limitation Iran, Syria, Sudan, Cuba, North Korea) and that he is
## not otherwise prohibited under the Export Laws from receiving the Software.
## All rights to Use the Software are granted on condition that such rights
## are forfeited if the Licensee fails to comply with
## the terms of this Agreement.
##
## 5.) SEVERABILITY
## Should any provision of this Agreement be or become invalid, ineffective
## or unenforceable, the remaining provisions of this Agreement shall be
## valid. The parties agree to replace the invalid, ineffective or
## unenforceable provision by a valid, effective and enforceable provision
## which best meets the commercial intention of the parties.
## The same shall apply in case of omissions.
##
## 6.) APPLICABLE LAW AND PLACE OF JURISTICATION
## This Agreement shall be constituted under the law of Germany.
## The United Nations Convention on the International Sale of Goods
## shall not apply to this Agreement. The Place of Jurisdiction for any
## dispute between the Parties shall be Tübingen (Germany).
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
## “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
## TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Narda Safety Test Solutions GmbH
## BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
## CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
## SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
## INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
## CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
## ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
## POSSIBILITY OF SUCH DAMAGE.
##
##############################################################################

from os import path
from pathlib import Path
import wx
import wx.grid as GD
import time
import datetime as dt
from dataclasses import dataclass
from operator import itemgetter, attrgetter
from PIL import ImageGrab, Image, ImageFile
import pyslipn.log as log
try:
    log = log.Log('')  # Disable logging
except AttributeError:
    # already set up, ignore error
    pass
from pyslipn import pyslipext
from pyslipn import map_local as mp
from nardascripting.base.usrscriptbase import *
from nardascripting.base.signalsharkdev import *
from nardascripting.base.measthread import *
import nardascripting.base.toolbox as tb
from . import scriptimages as ICO
from .cmapdlgsbase import CMapSettingsDlgBase, CMapDlgBase


def check_map_tiles(map_folder: str) -> bool:
    """Tries to find out if there is any slipy map data in the specified directory.

    :param map_folder: The folder to inspect.
    :return: Returns True if it is suspected to be a map folder.
    """
    result = False
    try:
        map_path = Path(map_folder)
        if map_path.exists():
            for sub_dir in map_path.iterdir():
                if sub_dir.is_dir() and sub_dir.stem.isdigit():
                    result = True
                    break
        return result
    except:
        return False


class LevelData(object):
    """Data class for level meter values"""

    def __init__(self):
        """Data class for level meter values"""
        self.timestamp = None
        self.gnss = None
        self.det_val = 0.0
        self.unit = "dBm"


class LevelScaleItem(object):
    """Data class for level scale items"""

    def __init__(self):
        """Data class for level scale items"""
        self.value = 0.0
        self.colour = wx.Colour(0, 0, 0)

    @staticmethod
    def from_values(value: float, colour: wx.Colour):
        """Generates LevelScaleItem from values"""
        retvalue = LevelScaleItem()
        retvalue.value = value
        retvalue.colour = colour
        return retvalue

    @staticmethod
    def from_dict(data: {}):
        """Generates LevelScaleItem from dict"""
        retvalue = LevelScaleItem()
        try:
            retvalue.value = data.get("value")
            retvalue.colour.Set(data.get("colour"))

        except:
            retvalue = LevelScaleItem()
        return retvalue

    def to_dict(self):
        """Generates dict from LevelScaleItem"""
        retvalue = {}
        retvalue["value"] = self.value
        retvalue["colour"] = self.colour.GetAsString(wx.C2S_HTML_SYNTAX)
        return retvalue

    def get_hex_color(self):
        return self.colour.GetAsString(wx.C2S_HTML_SYNTAX)


class LevelScale(object):
    """Class for level scale"""
    def __init__(self):
        """Class for level scale"""
        self.base_color = wx.Colour(51, 153, 51)
        self.items = []

    @staticmethod
    def get_default():
        """Generates default level scale"""
        retvalue = LevelScale()
        retvalue.add_item(LevelScaleItem.from_values(-112.0, wx.Colour(204, 51, 0)))
        retvalue.add_item(LevelScaleItem.from_values(-94.0, wx.Colour(255, 153, 0)))
        retvalue.add_item(LevelScaleItem.from_values(-76.0, wx.Colour(255, 255, 0)))
        retvalue.add_item(LevelScaleItem.from_values(-56.0, wx.Colour(204, 255, 51)))
        retvalue.add_item(LevelScaleItem.from_values(-40.0, wx.Colour(102, 255, 51)))
        return retvalue

    @staticmethod
    def from_dict(data: {}):
        """Generates LevelScale from dict"""
        retvalue = LevelScale()
        try:
            retvalue.base_color.Set(data.get("base_color"))
            level_items_dict = data.get("level_items")
            for key in level_items_dict:
                item = level_items_dict.get(key)
                retvalue.add_item(LevelScaleItem.from_dict(item))
            retvalue.items.sort(key=attrgetter("value"))

        except:
            retvalue = LevelScale.get_default()
        return retvalue

    def to_dict(self):
        """Generates dict from LevelScale"""
        retvalue = {}
        level_items_dict = {}
        for item in self.items:
            level_items_dict[str(item.value)] = item.to_dict()
        retvalue["level_items"] = level_items_dict
        retvalue["base_color"] = self.base_color.GetAsString(wx.C2S_HTML_SYNTAX)
        return retvalue

    def clear(self):
        """Clears all items"""
        self.items.clear()

    def add_item(self, item: LevelScaleItem):
        """Adds a LevelScaleItem to the Level Scale"""
        self.items.append(item)
        self.items.sort(key=attrgetter("value"))

    def __iter__(self):
        """Returns the Iterator object"""
        return iter(self.items)

    def get_last_item(self):
        """Returns the last LevelScaleItem or None"""
        last_index = len(self.items)-1
        if last_index >= 0:
            return self.items[last_index]
        return None

    def get_lev_color(self, level: float):
        """Calculates color for the given level"""
        for item in self.items:
            if level < item.value:
                return item.colour
        return self.base_color


@dataclass
class MeasParameters:
    """Class for keeping track of measurement parameters."""
    input: int = 1
    atten: float = 0.0
    unit: str = "dBm"
    ftune: float = 0.0
    cbw:float = 0.0
    det:str = "RMS"
    meas_time:float = 0.0
    post_avg: float = 0.0
    pre_avg: float = 0.0
    comment: str = ""

    def clone(self):
        """Clones the current data class"""
        retvalue = MeasParameters()
        retvalue.input = self.input
        retvalue.atten = self.atten
        retvalue.unit = self.unit
        retvalue.ftune = self.ftune
        retvalue.cbw = self.cbw
        retvalue.det = self.det
        retvalue.meas_time = self.meas_time
        retvalue.post_avg = self.post_avg
        retvalue.pre_avg = self.pre_avg
        retvalue.comment = self.comment
        return retvalue

    @staticmethod
    def from_csv_file(file_name: str):
        """Loads measurement parameters from csv file header"""
        meas_param = MeasParameters()
        delimiter = ";"
        try:
            with open(file_name, "r", newline="") as txt_reader:
                meas_param.input = int(txt_reader.readline().split(delimiter)[1])
                meas_param.atten = float(txt_reader.readline().split(delimiter)[1])
                meas_param.unit = txt_reader.readline().split(delimiter)[1].replace("\r\n", "")
                meas_param.ftune = float(txt_reader.readline().split(delimiter)[1])
                meas_param.cbw = float(txt_reader.readline().split(delimiter)[1])
                meas_param.det = txt_reader.readline().split(delimiter)[1].replace("\r\n", "")
                meas_param.meas_time = float(txt_reader.readline().split(delimiter)[1])
                meas_param.post_avg = float(txt_reader.readline().split(delimiter)[1])
                meas_param.pre_avg = float(txt_reader.readline().split(delimiter)[1])
                meas_param.comment = txt_reader.readline().split(delimiter)[1].replace("\r\n", "")
        except:
            meas_param = None
        return meas_param


class GridCellColorRenderer(GD.GridCellStringRenderer):
    """
    GridCellColorRenderer()

    The cell must contain a color string in the form of C2S_HTML_SYNTAX #"FF0000".
    """

    def __init__(self):
        """Default constructor"""
        super().__init__()


    def Draw(self, grid, attr, dc, rect, row, col, isSelected):
        """ Draw the given cell

        Draw the given cell on the provided DC inside the given rectangle using the style specified
        by the attribute and the default or selected state corresponding to the isSelected value.
        This pure virtual function has a default implementation which will prepare the DC using the given attribute:
        it will draw the rectangle with the background colour from attr and set the text colour and font.
        """
        bg = wx.Colour(255, 255, 255)
        text = grid.GetCellValue(row, col)
        try:
            bg = wx.Colour()
            bg.Set(text)
        except:
            bg = wx.Colour(255, 255, 255)

        #Defining the renderer What to draw
        cell_attr = attr.Clone()
        cell_attr.SetBackgroundColour(bg)
        super().Draw(grid, cell_attr, dc, rect, row, col, isSelected)


class CMapSettingsDlg(CMapSettingsDlgBase):
    """Custom dialog class for coverage map settings."""

    def __init__(self, parent, settings: dict):
        """Dialog class for data visualization.

        :param parent: The parent window as wx.Window.
        :param settings: Settings data as dictionary.
        """
        super().__init__(parent=parent)

        # Dynamic controls list
        self.controls = []

        # Level/color table base initialization:
        self.grd_level_scale.ClearGrid()
        self.grd_level_scale.InsertCols(numCols=4)
        self.grd_level_scale.EnableEditing(True)
        self.grd_level_scale.EnableGridLines(True)
        self.grd_level_scale.EnableDragGridSize(True)
        self.grd_level_scale.SetMargins(0, 0)
        self.grd_level_scale.SetDefaultRowSize(20)

        # Column rendering
        self.grd_level_scale.SetColLabelAlignment(wx.ALIGN_CENTRE, wx.ALIGN_CENTRE)

        cgray = (204, 204, 204)

        # Columns
        col = 0
        attr_design = GD.GridCellAttr()
        attr_design.SetReadOnly(True)
        attr_design.SetBackgroundColour(cgray)
        attr_design.SetRenderer(GD.GridCellStringRenderer())
        self.grd_level_scale.SetColSize(col, 20)
        self.grd_level_scale.SetColLabelValue(col, u" ")
        self.grd_level_scale.SetColAttr(col, attr_design)
        col += 1

        attr_float = GD.GridCellAttr()
        attr_float.SetReadOnly(False)
        attr_float.SetEditor(GD.GridCellFloatEditor(precision=1))
        attr_float.SetRenderer(GD.GridCellFloatRenderer(precision=1))
        self.grd_level_scale.SetColSize(col, 80)
        self.grd_level_scale.SetColLabelValue(col, u"Level")
        self.grd_level_scale.SetColAttr(col, attr_float)
        col += 1

        attr_unit = GD.GridCellAttr()
        attr_unit.SetReadOnly(True)
        attr_unit.SetBackgroundColour(cgray)
        attr_unit.SetRenderer(GD.GridCellStringRenderer())
        self.grd_level_scale.SetColSize(col, 60)
        self.grd_level_scale.SetColLabelValue(col, u"Unit")
        self.grd_level_scale.SetColAttr(col, attr_unit)
        col += 1

        attr_color = GD.GridCellAttr()
        attr_color.SetReadOnly(True)
        attr_color.SetRenderer(GridCellColorRenderer())
        self.grd_level_scale.SetColSize(col, 60)
        self.grd_level_scale.SetColLabelValue(col, u"Color")
        self.grd_level_scale.SetColAttr(col, attr_color)
        col += 1

        # load coverage map settings:
        self.settings = settings
        self.unit_str = ""
        self._apply_settings()

    def _apply_settings(self):
        """Applies the settings values to the GUI controls"""

        # Minimum point distance
        self.sctr_point_dist.SetValue(self.settings.get("point_dist", 500))

        # detector type:
        for item in LevDetectorModes:
            value = str(item)
            if value == "UNKNOWN" or value == "Off":
                continue
            self.cbx_detector.Append(value)
        self.cbx_detector.SetValue(self.settings.get("cbx_detector", "RMS"))

        # Unit
        for item in LevelUnits:
            value = str(item)
            if value == "UNKNOWN":
                continue
            self.cbx_unit.Append(value)
        self.unit_str = self.settings.get("unit", "dBm")
        self.cbx_unit.SetValue(self.unit_str)

        # File name prefix:
        self.tctr_file_prefix.SetValue(self.settings.get("tctr_file_prefix", "Measurement"))

        # File path:
        self.dirpck_file_path.SetPath(self.settings.get("dirpck_file_path", r"c:\\"))

        # Map path:
        self.dirpck_map_path.SetPath(self.settings.get("map_path", r"c:\\"))

        # Level scale:
        level_scale_dict = self.settings.get("level_scale", None)
        if level_scale_dict:
            self.level_scale = LevelScale.from_dict(level_scale_dict)
        else:
            self.level_scale = LevelScale.get_default()

        self.__init_level_scale_rows()

    def __init_level_scale_rows(self):
        """Initializes the level/color table rows"""
        num_rows = self.grd_level_scale.GetNumberRows()
        if num_rows > 0:
            self.grd_level_scale.DeleteRows(0, num_rows)

        for item in self.level_scale:
            self._add_color_legend_item("< ", item.value, self.unit_str, item.colour)
        self.cpkr_best_level_color.SetColour(self.level_scale.base_color)

    def _add_color_legend_item(self, designator: str, value: float, unit_str: str, colour: wx.Colour):
        """Creates a color legend row"""
        self.grd_level_scale.AppendRows(1)
        row_index = self.grd_level_scale.GetNumberRows() - 1
        self.grd_level_scale.SetCellValue(row_index, 0, designator)
        self.grd_level_scale.SetCellValue(row_index, 1, str(value))
        self.grd_level_scale.SetCellValue(row_index, 2, unit_str)
        self.grd_level_scale.SetCellValue(row_index, 3, colour.GetAsString(wx.C2S_HTML_SYNTAX))

    def on_btn_reset_color_scale(self, event):
        self.level_scale = LevelScale.get_default()
        self.__init_level_scale_rows()
        event.Skip()

    def _write_settings(self):
        """Writes the settings to the settings dictionary"""
        self.settings["point_dist"] = self.sctr_point_dist.GetValue()
        self.settings["cbx_detector"] = self.cbx_detector.GetValue()
        self.settings["unit"] = self.unit_str
        self.settings["tctr_file_prefix"] = self.tctr_file_prefix.GetValue()
        self.settings["dirpck_file_path"] = self.dirpck_file_path.GetPath()
        self.settings["map_path"] =self.dirpck_map_path.GetPath()

        self.level_scale.clear()
        self.level_scale.base_color = self.cpkr_best_level_color.GetColour()
        for row in range(self.grd_level_scale.NumberRows):
            new_item = LevelScaleItem()
            new_item.value = float(self.grd_level_scale.GetCellValue(row, 1))
            new_item.colour.Set(self.grd_level_scale.GetCellValue(row, 3))
            self.level_scale.add_item(new_item)
        self.settings["level_scale"] = self.level_scale.to_dict()

    def on_grid_cell_click(self, event):
        if event.Col != 3:
            event.Skip()
            return

        color_txt = self.grd_level_scale.GetCellValue(event.Row, event.Col)
        wx_color = wx.Colour(color_txt)
        with wx.ColourDialog(self) as dlg:
            color_data = dlg.GetColourData()
            color_data.SetCustomColour(0, wx_color)
            color_data.SetColour(wx_color)
            if dlg.ShowModal() == wx.ID_OK:
                color_txt_new = color_data.GetColour().GetAsString(wx.C2S_HTML_SYNTAX)
                self.grd_level_scale.SetCellValue(event.Row, event.Col, color_txt_new)

    def on_btn_delete_click(self, event):
        rows = self.grd_level_scale.GetSelectedRows()
        if rows:
            self.grd_level_scale.DeleteRows(rows[0])
        event.Skip()

    def on_btn_add_click(self, event):
        self._add_color_legend_item("<",0.0, self.unit_str, wx.Colour())
        event.Skip()

    def on_cbx_unit_changed(self, event):
        self.unit_str = self.cbx_unit.GetValue()
        for row in range(self.grd_level_scale.NumberRows):
            self.grd_level_scale.SetCellValue(row, 2, self.unit_str)
        event.Skip()

    def on_ok_button_click(self, event):
        self._write_settings()
        event.Skip()


class MapDlg(CMapDlgBase):
    """Custom dialog class for map visualization."""

    def __init__(self, parent, title: str, cmap_settings: dict, file_name: str,
                 meas_param: MeasParameters):
        """Dialog class for data visualization.

        :param parent: The parent window as wx.Window.
        :param cmap_settings: Settings for the map display.
        :param file_name: The file name.
        :param meas_param: Object that holds the measurement parameters.
        """
        super().__init__(parent=parent)

        self.cmap_settings = cmap_settings
        level_scale_dict = self.cmap_settings.get("level_scale", None)
        if level_scale_dict:
            self.level_scale = LevelScale.from_dict(level_scale_dict)
        else:
            self.level_scale = LevelScale.get_default()

        self.file_name = file_name
        self.txt_file_name.SetLabel(file_name)

        # Add images to the buttons:
        self.btn_zoom_in.SetBitmap(ICO.img_map_zoom_in.GetBitmap())
        self.btn_zoom_out.SetBitmap(ICO.img_map_zoom_out.GetBitmap())
        self.btn_center_pos.SetBitmap(ICO.img_map_center_pos.GetBitmap())

        # Set signal  label:
        self.txt_input.SetLabel("Input: {}".format(meas_param.input))
        self.txt_atten.SetLabel("Atten.: {} dB".format(meas_param.atten))
        self.txt_unit.SetLabel("Unit: {}".format(meas_param.unit))
        self.txt_ftune.SetLabel("Ftune: {} MHz".format(meas_param.ftune/1e6))
        self.txt_cbw.SetLabel("CBW: {} MHz".format(meas_param.cbw/1e6))
        self.txt_det.SetLabel("Det.: {}".format(meas_param.det))
        self.txt_meas_time.SetLabel("Meas. Time: {} s".format(meas_param.meas_time))
        self.txt_post_avg.SetLabel("Post Avg.: {} s".format(meas_param.post_avg))
        self.txt_pre_avg.SetLabel("Pre Avg.: {} s".format(meas_param.pre_avg))
        self.txt_comment.SetLabel(meas_param.comment)

        # Setup level values:
        unit_str = meas_param.unit

        # Create color legend
        self.controls = []
        self._point_colors = []
        for item in self.level_scale:
            self._add_color_legend_item(item.colour, "< {} {}".format(item.value, unit_str))
            self._point_colors.append(item.colour)
        last_item = self.level_scale.get_last_item()
        if last_item:
            self._add_color_legend_item(self.level_scale.base_color, ">= {} {}".format(last_item.value, unit_str))
            self._point_colors.append(self.level_scale.base_color)

        tiles_path = Path(cmap_settings.get("map_path", r"c:\\"))
        cr_text = "© unknown!"
        cr_file = tiles_path.joinpath("Copyright.txt")
        if cr_file.exists():
            cr_text = cr_file.read_text()
        self.txt_map_copyright.SetLabel(cr_text)

        self.first_run = True

        self.init_progress_bar(self.pgb_progress)
        self.init_dlg_buttons(self.sdbs_buttons, wx.OK)
        self.btn_ok.SetLabel("Quit")
        self.SetTitle(title)
        self.Maximize(True)

        # start application layout
        self.tile_source = mp.Tiles(tiles_path)
        self.pnl_map.SetBackgroundColour(wx.WHITE)
        self.pnl_map.ClearBackground()

        all_display = wx.BoxSizer(wx.HORIZONTAL)
        self.pnl_map.SetSizer(all_display)
        self.slip = pyslipext.pySlipExt(self.pnl_map, tile_src=self.tile_source, point_colors=self._point_colors,
                                        style=wx.SIMPLE_BORDER)

        self.data_layer_id = self.slip.AddPointLayer([], map_rel=True, visible=True,
                        name="data_points")

        all_display.Add(self.slip, proportion=1, flag=wx.EXPAND)
        self.pnl_map.SetSizerAndFit(all_display)

        self.last_pos = None

    def _add_color_legend_item(self, colour: wx.Colour, designator: str):
        """Creates a color legend row"""
        col_pan = wx.Panel(self.pnl_color_map, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, wx.TAB_TRAVERSAL)
        col_pan.SetBackgroundColour(colour)
        col_pan.SetMinSize(wx.Size(20, -1))
        self.controls.append(col_pan)
        self.fgs_color_map.Add(col_pan, 1, wx.EXPAND | wx.ALL, 5)

        txt_design = wx.StaticText(self.pnl_color_map, wx.ID_ANY, designator,
                                   wx.DefaultPosition, wx.DefaultSize, 0)
        txt_design.Wrap(-1)
        txt_design.SetFont(
            wx.Font(8, wx.FONTFAMILY_SWISS, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_NORMAL, False, "Arial"))
        self.controls.append(txt_design)
        self.fgs_color_map.Add(txt_design, 0, wx.ALL, 5)

    def save_screenshot(self):
        """ Saves current screen as png.

        :return:
        """
        try:
            meas_file = Path(self.file_name)
            initdir = str(meas_file.parent)
            time_stamp = dt.datetime.now()

            default_file_name = time_stamp.strftime(meas_file.stem + "_%Y-%m-%d_%H-%M-%S.png")
            wildcard = "PNG file (*.png)|*.png"
            filename = ""
            with wx.FileDialog(self, message="Save screenshot to", defaultDir=initdir,
                               defaultFile=default_file_name, wildcard=wildcard, style=wx.FD_SAVE) as dlg:
                if dlg.ShowModal() == wx.ID_OK:
                    filename = dlg.GetPath()
                    screen_rect = self.GetScreenRect()
                    left_x, top_y = screen_rect.GetTopLeft().Get()
                    right_x, bottom_y = screen_rect.GetBottomRight().Get()
                    screenshot = ImageGrab.grab(bbox=(left_x, top_y, right_x, bottom_y))
                    screenshot.save(filename, 'PNG')
                    self.lbl_status.SetLabelText("Screenshot saved.")
        except:
            pass

    def on_zoom_in_clicked(self, event):
        level, geo = self.slip.GetLevelAndPosition()
        self.slip.GotoLevelAndPosition(level + 1, geo)

    def on_zoom_out_clicked(self, event):
        level, geo = self.slip.GetLevelAndPosition()
        self.slip.GotoLevelAndPosition(level - 1, geo)

    def on_center_pos_clicked(self, event):
        if self.last_pos:
            self.slip.GotoPosition(self.last_pos)

    def on_btn_save_png_click(self, event):
        self.save_screenshot()

    def on_close(self, event):
        self.slip.del_dc_points()
        event.Skip()

    def update_progress(self, evt: ProgressEvent):
        """Receives data from thread and updates progressbar and message text

        :param evt: Event of type ProgressEvent.
        """
        # Handle progress
        self.update_progress_bar(evt, self.pgb_progress)

        # Handle message
        if evt.msg is not None:
            self.lbl_status.SetLabelText(evt.msg)

        # Handle data:
        if evt.data is not None:
            if self.first_run:
                self.slip.GotoLevel(12)
                self.slip.GotoPosition(evt.data[0].gnss)
                self.first_run = False

            # Add data point(s)
            points = []
            data_cnt = len(evt.data)

            for item in evt.data:
                attributes = {"placement": "cc", "radius": 5, "colour": self.level_scale.get_lev_color(item.det_val), "offset_x": 0,
                          "offset_y": 0, "data": None}
                points.append((item.gnss.longitude, item.gnss.latitude, attributes))
            self.slip.add_points(self.data_layer_id, points)
            self.last_pos = evt.data[data_cnt-1].gnss

        # Handle buttons:
        if evt.btn_style:
            self.update_btn_style(evt.btn_style)

        # Handle icons:
        if evt.icon_style:
            # self.update_icon_style(self.bmp_icon, evt.icon_style)
            pass

        # Handle help text:
        if evt.help_text:
            self.help_text = evt.help_text

        self.Layout()
        self.Refresh()
        self.Update()


# This is just a python module containing helper classes/functions:
if __name__ == '__main__':
    pass